/*
 * Copyright (C) 2001-2010 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef DCPLUSPLUS_DCPP_HASH_MANAGER_H
#define DCPLUSPLUS_DCPP_HASH_MANAGER_H

#include "Singleton.h"
#include "MerkleTree.h"
#include "Thread.h"
#include "Semaphore.h"
#include "TimerManager.h"
#include "Util.h"
#include "FastAlloc.h"
#include "Text.h"
#include "Streams.h"
#include "CFlyMediaInfo.h"

#ifdef RIP_USE_STREAM_SUPPORT_DETECTION
#include "FsUtils.h"
#endif

#define IRAINMAN_NTFS_STREAM_TTH

#define MAX_HASH_PROGRESS_VALUE 3000 //[+] brain-ripper

STANDARD_EXCEPTION(HashException);
class File;

class HashManagerListener
{
	public:
		virtual ~HashManagerListener() { }
		template<int I> struct X
		{
			enum { TYPE = I };
		};
		
		typedef X<0> TTHDone;
		
		virtual void on(TTHDone, const string& /* fileName */, const TTHValue& /* root */,
		                int64_t aTimeStamp, const CFlyMediaInfo& p_media) throw() = 0;
};

class FileException;

class HashManager : public Singleton<HashManager>, public Speaker<HashManagerListener>,
	private TimerManagerListener
{
	public:
	
		HashManager()
		{
			TimerManager::getInstance()->addListener(this);
		}
		virtual ~HashManager() throw()
		{
			TimerManager::getInstance()->removeListener(this);
			hasher.join();
		}
		
		inline void hashFile(const string& fileName, int64_t aSize)
		{
			Lock l(cs);
			hasher.hashFile(fileName, aSize);
		}
		
		/**
		 * Check if the TTH tree associated with the filename is current.
		 */
		bool checkTTH(const string& fname, const string& fpath, int64_t p_path_id, int64_t aSize, int64_t aTimeStamp, TTHValue& p_out_tth);
		void stopHashing(const string& baseDir)
		{
			hasher.stopHashing(baseDir);
		}
		inline void setPriority(Thread::Priority p)
		{
			hasher.setThreadPriority(p);
		}
		
		/** Return block size of the tree associated with root, or 0 if no such tree is in the store */
		size_t getBlockSize(const TTHValue& root);
		
		int64_t getBlockSize(int64_t p_file_size)
		{
			const int64_t l_result = max(TigerTree::calcBlockSize(p_file_size, 10), MIN_BLOCK_SIZE);
			return l_result;
		}
		/** @return TTH root */
		const TTHValue getTTH(const string& fname, const string& fpath, int64_t aSize) throw(HashException);
		
		inline void addTree(const string& aFileName, uint64_t aTimeStamp, const TigerTree& tt, const CFlyMediaInfo& p_media)
		{
			hashDone(aFileName, aTimeStamp, tt, -1, p_media, false);
		}
		void addTree(const TigerTree& p_tree);
		
		inline void getStats(string& curFile, int64_t& bytesLeft, size_t& filesLeft)
		{
			hasher.getStats(curFile, bytesLeft, filesLeft);
		}
		
		/**
		 * Rebuild hash data file
		 */
		void rebuild()
		{
			hasher.scheduleRebuild();
		}
		
		void startup()
		{
			hasher.start();
		}
		
		void shutdown()
		{
			// [-] brain-ripper
			// Removed critical section - it caused deadlock on client exit:
			// "join" hangs on WaitForSingleObject(threadHandle, INFINITE),
			// and threadHandle thread hangs on this very critical section
			// in HashManager::hashDone
			//Lock l(cs); //[!]IRainman
			
			hasher.shutdown();
			hasher.join();
		}
		void getMediaInfo(const string& p_name, CFlyMediaInfo& p_media);// [+]IRainman
		
		struct HashPauser
		{
				HashPauser();
				~HashPauser();
				
			private:
				bool resume;
		};
		
		/// @return whether hashing was already paused
		bool pauseHashing();
		void resumeHashing();
		bool isHashingPaused() const;
		
		
//[+] Greylink
#ifdef IRAINMAN_NTFS_STREAM_TTH
		class StreamStore   // greylink dc++: work with ntfs stream
		{
			public:
				bool loadTree(const string& p_filePath, TigerTree& p_Tree, int64_t p_aFileSize = -1);
				bool saveTree(const string& p_filePath, const TigerTree& p_Tree);// [+] IRainman const string& p_filePath
				void deleteStream(const string& p_filePath);//[+] IRainman
				
#ifdef RIP_USE_STREAM_SUPPORT_DETECTION
				void SetFsDetectorNotifyWnd(HWND hWnd);
				void OnDeviceChange(LPARAM lParam, WPARAM wParam)
				{
					m_FsDetect.OnDeviceChange(lParam, wParam);
				}
#endif
			private:
				struct TTHStreamHeader
				{
					uint32_t magic;
					uint32_t checksum;  // xor of other TTHStreamHeader DWORDs
					uint64_t fileSize;
					uint64_t timeStamp;
					uint64_t blockSize;
					TTHValue root;
				};
//              static const uint64_t g_minStreamedFileSize = 16*1048576;//[-]NightOrion
				static const uint32_t g_MAGIC = '++lg';
				static const string g_streamName;
#ifndef RIP_USE_STREAM_SUPPORT_DETECTION
				static std::set<char> g_error_tth_stream; //[+]PPA
				static void addBan(const string& p_filePath)
				{
					if (p_filePath.size())
						g_error_tth_stream.insert(p_filePath[0]);
				}
				static bool isBan(const string& p_filePath)
				{
					return p_filePath.size() && g_error_tth_stream.count(p_filePath[0]) == 1;
				}
#endif
				
				void setCheckSum(TTHStreamHeader& p_header);
				bool validateCheckSum(const TTHStreamHeader& p_header);
#ifdef RIP_USE_STREAM_SUPPORT_DETECTION
				// [+] brain-ripper
				// Detector if volume support streams
				FsUtils::CFsTypeDetector m_FsDetect;
#endif
		};
//[~] Greylink
		StreamStore m_streamstore;
#endif // IRAINMAN_NTFS_STREAM_TTH
		
#ifdef RIP_USE_STREAM_SUPPORT_DETECTION
		void SetFsDetectorNotifyWnd(HWND hWnd);
		void OnDeviceChange(LPARAM lParam, WPARAM wParam)
		{
			m_streamstore.OnDeviceChange(lParam, wParam);
		}
#endif
		
		// [+] brain-ripper
		// Temporarily change hash speed functional
		inline bool IsHashing()
		{
			return hasher.IsHashing();
		}
		inline size_t GetProgressValue()
		{
			return hasher.GetProgressValue();
		}
		inline uint64_t GetStartTime()
		{
			return hasher.GetStartTime();
		}
		inline void GetProcessedFilesAndBytesCount(int64_t &bytes, size_t &files)
		{
			hasher.GetProcessedFilesAndBytesCount(bytes, files);
		}
		inline DWORD GetMaxProgressValue()
		{
			return hasher.GetMaxProgressValue();
		}
		inline void EnableForceMinHashSpeed(int iMinHashSpeed)
		{
			hasher.EnableForceMinHashSpeed(iMinHashSpeed);
		}
		inline void DisableForceMinHashSpeed()
		{
			hasher.DisableForceMinHashSpeed();
		}
		inline int GetMaxHashSpeed()
		{
			return hasher.GetMaxHashSpeed();
		}
		// [~] brain-ripper
		
		
	private:
		class Hasher : public Thread
		{
			public:
				Hasher() : stop(false), running(false), paused(0), rebuild(false), currentSize(0),
					m_CurrentBytesLeft(0), //[+]IRainman
					bForceMaxHashSpeed(false), iForceMaxHashSpeed(0), dwMaxFiles(0), iMaxBytes(0) { }
					
				void hashFile(const string& fileName, int64_t size);
				
				/// @return whether hashing was already paused
				bool pause();
				void resume();
				bool isPaused() const;
				
				void stopHashing(const string& baseDir);
				int run();
				bool fastHash(const string& fname, uint8_t* buf, TigerTree& tth, int64_t size);
				void getStats(string& curFile, int64_t& bytesLeft, size_t& filesLeft);
				void shutdown()
				{
					stop = true;
					if (paused) s.signal();
					s.signal();
				}
				void scheduleRebuild()
				{
					rebuild = true;
					if (paused) s.signal();
					s.signal();
				}
				
				int GetMaxHashSpeed();
				
				// [+] brain-ripper
				// Temporarily change hash speed functional
				void HashManager::Hasher::getBytesAndFileLeft(int64_t& bytesLeft, size_t& filesLeft);
				inline void EnableForceMinHashSpeed(int iMinHashSpeed)
				{
					bForceMaxHashSpeed = true;
					iForceMaxHashSpeed = iMinHashSpeed;
				}
				inline void DisableForceMinHashSpeed()
				{
					bForceMaxHashSpeed = false;
					iForceMaxHashSpeed = 0;
				}
				
				inline bool IsHashing()
				{
					return running;
				}
				
				inline size_t GetProgressValue()
				{
					if (!running || dwMaxFiles == 0 || iMaxBytes == 0)
						return 0;
						
					int64_t bytesLeft;
					size_t filesLeft;
					getBytesAndFileLeft(bytesLeft, filesLeft);
					
					return static_cast<size_t>(MAX_HASH_PROGRESS_VALUE * ((0.5 * static_cast<double>(dwMaxFiles - filesLeft) / static_cast<double>(dwMaxFiles)) + 0.5 * static_cast<double>(iMaxBytes - bytesLeft) / static_cast<double>(iMaxBytes)));
				}
				
				inline uint64_t GetStartTime()
				{
					return IsHashing() ? uiStartTime : 0;
				}
				
				inline void GetProcessedFilesAndBytesCount(int64_t &bytes, size_t &files)
				{
					if (IsHashing())
					{
						int64_t bytesLeft;
						size_t filesLeft;
						getBytesAndFileLeft(bytesLeft, filesLeft);
						bytes = iMaxBytes >= bytesLeft ? iMaxBytes - bytesLeft : iMaxBytes;
						files = dwMaxFiles >= filesLeft ? dwMaxFiles - filesLeft : dwMaxFiles;
					}
					else
					{
						bytes = 0;
						files = 0;
					}
				}
				
				inline DWORD GetMaxProgressValue()
				{
					return MAX_HASH_PROGRESS_VALUE;
				}
				// [+] brain-ripper
				// end of Temporarily change hash speed functional
				
			private:
				// Case-sensitive (faster), it is rather unlikely that case changes, and if it does it's harmless.
				// map because it's sorted (to avoid random hash order that would create quite strange shares while hashing)
				typedef map<string, int64_t> WorkMap;
				typedef WorkMap::iterator WorkIter;
				
				WorkMap w;
				mutable CriticalSection cs;
				Semaphore s;
				
				bool stop;
				bool running;
				int64_t paused; //[!] PPA -> int
				bool rebuild;
				//string currentFile;// [-]IRainman
				int64_t currentSize;
				
				void instantPause();
				
				// [+] brain-ripper
				bool bForceMaxHashSpeed;
				int iForceMaxHashSpeed;
				size_t dwMaxFiles;
				int64_t iMaxBytes;
				uint64_t uiStartTime;
				// -----------------
				
				// [+]IRainman for better performance
				int64_t m_CurrentBytesLeft;
				string m_fname;
				// [~]IRainman
		};
		
		friend class Hasher;
		
		void addFile(const string& aFileName, int64_t aTimeStamp, const TigerTree& tth, const CFlyMediaInfo& p_media);
#ifdef IRAINMAN_NTFS_STREAM_TTH
		void HashManager::addFileFromStream(const string& p_name, const TigerTree& p_TT);
#endif
		
		Hasher hasher;
		
		mutable CriticalSection cs;
		
		/** Single node tree where node = root, no storage in HashData.dat */
		static const int64_t SMALL_TREE = -1;
		
		void hashDone(const string& aFileName, uint64_t aTimeStamp, const TigerTree& tth, int64_t speed,
		              const CFlyMediaInfo& p_media,
		              bool p_is_ntfs);
		void doRebuild()
		{
			Lock l(cs);
			rebuild();
		}
};

#endif // !defined(HASH_MANAGER_H)

/**
 * @file
 * $Id: HashManager.h 482 2010-02-13 10:49:30Z bigmuscle $
 */
